
    #   Assemble canon of transits from list of transit
    #	events in the numerical integration.
    
    #	Input:	all_fine.txt.gz
    #	Output: transits.csv

    #	We read the output from transits.c, which lists all
    #	transits in progress for each step of the numerical
    #	integration, and prepare a table of transit events,
    #	listing the start, maximum, and end of each.  Note
    #	that these are not the actual times of the transit
    #	events but rather the integration step time which
    #	closest to the actual event.
    
    #	by John Walker
    #	http://www.fourmilab.ch/
    
    #	For details, please see: http://www.fourmilab.ch/documents/canon_transits/
    
    use strict;
    use warnings;
    
    my $trace = 0;  	    	    # Trace event evolution ?
    my $progress = 1;	    	    # Show progress ?
    
    my $l;
    my @rtransits;  	    	    # Unsorted transit records
    
    my $ljd = 0;
    my $lpd = -1E100;	    	    # Last progress century printed
    my $ltr = 0;    	    	    # Last transit count
    
    my ($of, $from, $jd, $details, $wday, $year, $mon, $day, $hour, $min, $sunrad, $objang, $offset);
    my (%trans_start, %trans_first, %trans_maxd, %trans_last, %trans_closest, %trans_max, %trans_end);
    
    open(FI, "zcat ../Integration_Results/all_fine.txt.gz |") ||
    	die("Cannot open pipe from ../Integration_Results/all_fine.txt.gz");
    
    while ($l = <FI>) {
    
	if (!parse_transit($l)) {
    	    next;
	}
print("**** $l") if $trace;

    	if ($progress) {
	    my $pc = int($year / 1000);
	    if ($pc > $lpd) {
	    	my $tc = ($#rtransits > 0) ? $#rtransits : 0;
		my $tpy = int(($tc - $ltr) / 1000);
	    	printf(STDERR "... %d-%02d-%02d ...  %d transits, %d per year\n", $year, $mon, $day, $#rtransits, $tpy);
		$ltr = $tc;
	    	$lpd = $pc;
	    }
	}

    	#   If this item has a different Julian Day than the last, we've
	#   now seen every event for day $ljd.  Now we walk through the list
	#   of open events and close any which didn't post a continuation
	#   in the last day's bunch.
	
	if ($jd != $ljd) {
	    check_trans_end($ljd, $jd);
	    parse_transit($l);
	    $ljd = $jd;
	}
	
	#   We identify a transit event by its $of and $from
	#   objects.  Form this identity, then test whether
	#   a transit is presently underway.  If not, open
	#   a new transit event.  Note that we begin by setting the
	#   maximum event and time to the first one seen, since if
	#   this is a near-graze which we only see in one step,
	#   the first will be the maximum.
	
	my $transid = "$of-$from";
	if ((!defined($trans_start{$transid})) || ($trans_start{$transid} eq '')) {
	    $trans_start{$transid} = $l;    # Start of transit
	    $trans_first{$transid} = $jd;
	    $trans_closest{$transid} = $offset;
	    $trans_max{$transid} = $l;
	    $trans_maxd{$transid} = $jd;
print("Opening $transid event at $jd\n") if $trace;
	}
	
	#   If this point in the transit is closer to the centre
	#   of the Sun than the previous, save this as a candidate
	#   closest time.
	
print("Testing $transid closer to $offset than $trans_closest{$transid} at $jd\n") if $trace;
	if ($offset > $trans_closest{$transid}) {
	    $trans_closest{$transid} = $offset;
	    $trans_max{$transid} = $l;
	    $trans_maxd{$transid} = $jd;
print("Updating $transid closest to $offset at $jd\n") if $trace;
	}
	
	#   Save the most recent Julian day of the transit in progress
	#   and the current record as a candidate for end of transit.
	
	$trans_last{$transid} = $jd;    # Most recent in-transit time
	$trans_end{$transid} = $l;
print("Continuing $transid event at $trans_last{$transid}\n") if $trace;
    }
    
    #	Flush out transits pending at end of file.  Note that
    #	in most cases (but not necessarily), any such transits will
    #	have ended, but simply not have been terminated because
    #	no subsequent transit was found in later steps of the
    #	numerical integration.
    
    $jd += 1;
    check_trans_end($ljd, $ljd + 100);
    
    close(FI);
    
    #	Sort the transits by start time
    
    print(STDERR "Sorting transits by start time.\n");
    my @transits = sort bystart @rtransits;
    undef @rtransits;
    
    print(STDERR "Writing transits to transits.csv.\n");
    open(FO, ">transits.csv") || die("Cannot write transits.csv");
	for (my $i = 0; $i <= $#transits; $i++) {
    	    print(FO $transits[$i]);
	}
    close(FO);
    
    #	When the Julian date changes, check for any transits which
    #	have ended.  Such a transit will be present in %trans_start
    #	table but have a %trans_last different than the Julian date
    #	of the present event.  We also have to cope with the case
    #	where we see the last event of a given transit, say a transit
    #	of Mercury from Uranus, and we don't see any other transits
    #	until the next transit of Mercury from Uranus, 88-odd days
    #	later.  If we just compared the last Julian date, we'd mistakenly
    #	report the transit as one continuous event.  So, we simply test
    #	whether there's a gap of one or more days between the current
    #	event date and the last one.  This is far longer than our
    #	integration step and much shorter than minimum interval
    #	between pairs of transits of the same planets.  If this is
    #	the case, we close out the old event on the basis of the gap.

    sub check_trans_end {
    	my ($ljd, $jd) = @_;
    	my $k;
	
    	foreach $k (keys %trans_start) {
	    if ($trans_start{$k} ne '') {
		if (($trans_last{$k} != $ljd) || (($jd - ($trans_last{$k})) > 1)) {
#print("++ $k $trans_last{$k}  $ljd\n");
print("Closing $k event at $ljd  trans_last = $trans_last{$k}\n") if $trace;
	    	    parse_transit($trans_start{$k});
		    
    	    	    my ($start, $mid, $end, $subsubtends, $transitangle);
	    	    $start = trimTrailingZeroes($trans_first{$k});
	    	    $mid = trimTrailingZeroes($trans_maxd{$k});
	    	    $end = trimTrailingZeroes($trans_last{$k});
		    parse_transit($trans_max{$k});
		    $subsubtends = trimTrailingZeroes($sunrad);
		    $transitangle = trimTrailingZeroes($objang);
		    $subsubtends = trimZeroIntegerPart($subsubtends);
		    $transitangle = trimZeroIntegerPart($transitangle);
	    	    push(@rtransits, "$of,$from,$start,$mid,$end,$subsubtends,$transitangle\n");

		    $trans_start{$k} = '';
		}
else {
print("Pursuing $k event at $ljd $trans_last{$k}\n") if $trace;
}
	    }
	}
    }
    
    #	Comparison function to sort transit records by start time
    
    sub bystart {
	$a =~ m/^[^\,]+,[^\,]+,([^\,]+),/;
    	my $astart = $1;
	$b =~ m/^[^\,]+,[^\,]+,([^\,]+),/;
    	my $bstart = $1;
	return $astart <=> $bstart;
    }    
    
    #	Trim trailing zeroes from the decimal part of a floating point number

    sub trimTrailingZeroes {
    	my $n = shift();
	
	if ($n =~ m/\.\d+$/) {	    # Make sure there's a decimal point
	    $n =~ s/0+$//;
	}
	#   Remove trailing decimal if there are preceding digits, but
	#   don't leave the number as just '.'.
	if ($n eq '.') {
	    $n = '0';
    	} elsif ($n =~ m/\.$/) {
	    $n =~ s/\.$//;
	}
	return $n;
    }
    
    #	Eliminate zero integer part from a floating point number
    
    sub trimZeroIntegerPart {
    	my $n = shift();
    	
	$n =~ s/^(\-?)0\./$1\./;
	if ($n eq '.') {
	    $n = '0';
	}
	return $n;
    }

    #	Parse a transit event record into global field variables
    
    sub parse_transit {
	my ($l) = @_;
	
    	$l =~ s/\s+$//;
	if ($l =~ m/^Transit of (\w+)\s+from (\w+)\s+(\S+)\s+(.*)$/) {
	    ($of, $from, $jd, $details) = ($1, $2, $3, $4);
	    if (!($details =~ m/(\w+)\s+([\-\d]+)\-(\d+)\-(\d+)\s+(\d+):(\d+)\s+([\d\.]+)\s+([\d\.]+)\s+([\d\.]+)/)) {
	    	die("Failed to parse details: \"$details\"\n");
	    }
	    ($wday, $year, $mon, $day, $hour, $min, $sunrad, $objang, $offset) =
		($1, $2, $3, $4, $5, $6, $7, $8, $9);
		
	    return 1;
	}
	return 0;
    }
